import warnings


class jpeWarning(Warning):pass

class jpeDevFuncWarning(jpeWarning):
    " warning raised when function is jut for my benifit and isnt in the doc"
    pass

class jpeClassDataChange(jpeWarning):
    " when a function significantaly changes a class without good reason"
    pass

class jpeRedundantCall(jpeWarning):
    " called when a redundant things are done"
    pass



"""Perlin noise implementation."""
# Licensed under ISC
from itertools import product
import math
import random


def smoothstep(t):
    """Smooth curve with a zero derivative at 0 and 1, making it useful for
    interpolating.
    """
    return t * t * (3. - 2. * t)


def lerp(t, a, b):
    """Linear interpolation between a and b, given a fraction t."""
    return a + t * (b - a)


class Factory(object):
    """Callable that produces Perlin noise for an arbitrary point in an
    arbitrary number of dimensions.  The underlying grid is aligned with the
    integers.
    There is no limit to the coordinates used; new gradients are generated on
    the fly as necessary.
    """

    def __init__(self, dimension, octaves=1, tile=(), unbias=False):
        """Create a new Perlin noise factory in the given number of dimensions,
        which should be an integer and at least 1.
        More octaves create a foggier and more-detailed noise pattern.  More
        than 4 octaves is rather excessive.
        ``tile`` can be used to make a seamlessly tiling pattern.  For example:
            pnf = PerlinNoiseFactory(2, tile=(0, 3))
        This will produce noise that tiles every 3 units vertically, but never
        tiles horizontally.
        If ``unbias`` is true, the smoothstep function will be applied to the
        output before returning it, to counteract some of Perlin noise's
        significant bias towards the center of its output range.
        """
        self.dimension = dimension
        self.octaves = octaves
        self.tile = tile + (0,) * dimension
        self.unbias = unbias

        # For n dimensions, the range of Perlin noise is ±sqrt(n)/2; multiply
        # by this to scale to ±1
        self.scale_factor = 2 * dimension ** -0.5

        self.gradient = {}

    def _generate_gradient(self):
        # Generate a random unit vector at each grid point -- this is the
        # "gradient" vector, in that the grid tile slopes towards it

        # 1 dimension is special, since the only unit vector is trivial;
        # instead, use a slope between -1 and 1
        if self.dimension == 1:
            return (random.uniform(-1, 1),)

        # Generate a random point on the surface of the unit n-hypersphere;
        # this is the same as a random unit vector in n dimensions.  Thanks
        # to: http://mathworld.wolfram.com/SpherePointPicking.html
        # Pick n normal random variables with stddev 1
        random_point = [random.gauss(0, 1) for _ in range(self.dimension)]
        # Then scale the result to a unit vector
        scale = sum(n * n for n in random_point) ** -0.5
        return tuple(coord * scale for coord in random_point)

    def get_plain_noise(self, *point):
        """Get plain noise for a single point, without taking into account
        either octaves or tiling.
        """
        if len(point) != self.dimension:
            raise ValueError("Expected {} values, got {}".format(
                self.dimension, len(point)))

        # Build a list of the (min, max) bounds in each dimension
        grid_coords = []
        for coord in point:
            min_coord = math.floor(coord)
            max_coord = min_coord + 1
            grid_coords.append((min_coord, max_coord))

        # Compute the dot product of each gradient vector and the point's
        # distance from the corresponding grid point.  This gives you each
        # gradient's "influence" on the chosen point.
        dots = []
        for grid_point in product(*grid_coords):
            if grid_point not in self.gradient:
                self.gradient[grid_point] = self._generate_gradient()
            gradient = self.gradient[grid_point]

            dot = 0
            for i in range(self.dimension):
                dot += gradient[i] * (point[i] - grid_point[i])
            dots.append(dot)

        # Interpolate all those dot products together.  The interpolation is
        # done with smoothstep to smooth out the slope as you pass from one
        # grid cell into the next.
        # Due to the way product() works, dot products are ordered such that
        # the last dimension alternates: (..., min), (..., max), etc.  So we
        # can interpolate adjacent pairs to "collapse" that last dimension.  Then
        # the results will alternate in their second-to-last dimension, and so
        # forth, until we only have a single value left.
        dim = self.dimension
        while len(dots) > 1:
            dim -= 1
            s = smoothstep(point[dim] - grid_coords[dim][0])

            next_dots = []
            while dots:
                next_dots.append(lerp(s, dots.pop(0), dots.pop(0)))

            dots = next_dots

        return dots[0] * self.scale_factor

    def __call__(self, *point):
        """Get the value of this Perlin noise function at the given point.  The
        number of values given should match the number of dimensions.
        """
        ret = 0
        for o in range(self.octaves):
            o2 = 1 << o
            new_point = []
            for i, coord in enumerate(point):
                coord *= o2
                if self.tile[i]:
                    coord %= self.tile[i] * o2
                new_point.append(coord)
            ret += self.get_plain_noise(*new_point) / o2

        # Need to scale n back down since adding all those extra octaves has
        # probably expanded it beyond ±1
        # 1 octave: ±1
        # 2 octaves: ±1½
        # 3 octaves: ±1¾
        ret /= 2 - 2 ** (1 - self.octaves)

        if self.unbias:
            # The output of the plain Perlin noise algorithm has a fairly
            # strong bias towards the center due to the central limit theorem
            # -- in fact the top and bottom 1/8 virtually never happen.  That's
            # a quarter of our entire output range!  If only we had a function
            # in [0..1] that could introduce a bias towards the endpoints...
            r = (ret + 1) / 2
            # Doing it this many times is a completely made-up heuristic.
            for _ in range(int(self.octaves / 2 + 0.5)):
                r = smoothstep(r)
            ret = r * 2 - 1

        return ret



import time

class timer:
    def __init__(self, acc=1):
        self.laps = []
        self.startTime = None
        self.acc = 1

    def start(self):
        self.startTime = time.time()
        self.clear()

    def clear(self):
        self.laps = []

    def lap(self, msg):
        self.endLap(msg)

    def endLap(self, msg):
        if self.startTime is None:
            raise TimerNotStartedError("u need to start the timer dumbo")
        dt = time.time() - self.startTime
        self.laps.append((dt, msg))
        self.startTime = time.time()

    def end(self, msg='last'):
        self.endLap(msg)
        self.startTime = None
        print(self)

    def __str__(self):
        output = ""
        for lap in self.laps:
            output += (f"lap {lap[1]} ended and took {round(lap[0], self.acc)} sec\n")
        return output[:-1]


class TimerNotStartedError(Exception):
    pass



import numpy as np
import jpe.Jfloat
import math

def jVectorTypeErrorMessage(other, operation):
    raise TypeError(f"jpe.vector dosnt suport {operation} operation with values of type {type(other)} if u think this schould be the case plz write me a message under Julian.wandhoven@fgz.ch")

def dot(u, v):
    return u._arr.dot(v._arr)

def cross(u, v):
    return vector(np.cross(u._arr, v._arr))

def getAngle(u, v):
    return math.acos(dot(u, v) /(v.__abs__() * u.__abs__()))

def det(v):
    return np.linalg.det(v)

def resultingForce(*vectors):
    try:
        vec = vectors[0]
        for cvec in vectors[1:]:
            vec = vec + cvec
    except:
        raise ValueError("dimentions do not match")
    return vec
    


class vector():
    def __init__(self, *val):
        if isinstance(val[0], np.ndarray):
            self._arr = val[0]
        else: self._arr = np.array(val)
        if (len(self._arr.shape)) != 1:
            raise ValueError(f"cant convert {self._arr} to vector")
        
    def __str__(self):
        return f"vector ({self._arr.__str__()[1:-1]})"
# addition
    def __add__(self, other):
        if isinstance(other, vector):
            return vector(self._arr + other._arr)
        jVectorTypeErrorMessage(other, "addition")

    def __radd__(self, other): return self.__add__(other)


#subraction
    def __sub__(self, other):
        if isinstance(other, vector):
            return vector(self._arr - other._arr)
        jVectorTypeErrorMessage(other, "diverentiation")

    def __rsub__(self, other):
        if isinstance(other, vector):
            return vector(other._arr - self._arr)
        jVectorTypeErrorMessage(other, "diverentiation")

    
#scalar
        #mul
    def __mul__(self, other):
        if isinstance(other, int) or isinstance(other, float) or isinstance(jpe.Jfloat.Jfloat):
            return vector(self._arr * other)
        else:
            jVectorTypeErrorMessage(other, "scalar")

    def __rmul__(self, other): return self.__mul__(other)
    
    
    #div    
    def __truediv__(self, other):
        if isinstance(other, int) or isinstance(other, float) or isinstance(jpe.Jfloat.Jfloat):
            return vector(self._arr / other)
        else:
            jVectorTypeErrorMessage(other, "scalar")

    def __floordiv__(self, other):
        if isinstance(other, int) or isinstance(other, float) or isinstance(jpe.Jfloat.Jfloat):
            return vector(self._arr // other)
        else:
            jVectorTypeErrorMessage(other, "floor scalar")

#lengths
    def __abs__(self):
        return np.linalg.norm(self._arr)

    def unit(self):
        return vector(self._arr / self.__abs__())

#muls
    def dot(self, other):
        return dot(self, other)

    def cross(self, other):
        return cross(self, other)

    def getAngle(s, o):
        return getAngle(s, o)

    def det(self):
        return det(self)
    
    def __neg__(self):
        return vector(-self._arr)

    def __getitem__(self, idx):
        return self._arr[idx]

    def __repr__(self): return self.__str__()
        



import math
import jpe.math.Jmath as Jmath

class Jfloat():
    
    def __init__(self, val, N=None): ## input val as 123.456|678 | commences periode

        if N is None:
            self.n, self.x = self._getCo(val)
        else:
            self.n, self.x = self._simplify((N, val))

        ##float.__init__(self, self.x / self.n)


    def _getCo(self, val):
        return self._simplify(self._getCo_2(val))

    def _simplify(self, co):
        n, x = co
        gcd = math.gcd(x, n)
        if n < 0: x, n = -x, -n
        return (n//gcd, x//gcd)
        
    
    def _getCo_2(self, val):
        if isinstance(val, str):
            return self._getCoFromStr(val)
        if isinstance(val, int):
            return 1, val
        if isinstance(val, float):
            return self._getCoFromStr(str(val))
        raise ValueError ('invalid literal for Jfloat(): ' + str(val))

    def _getCoFromStr(self, val):
        if not '.' in val:
            return 1, int(val)
        
        if not '|' in val:
            return self._getCoFromStr_simple(val)
        return self._getCoFromStr_wPer(val)

    def _getCoFromStr_simple(self, val):
        dot_idx = val.index('.')

        n = int(10 ** (len(val) - dot_idx- 1))
        x = int(val.replace('.', ''))
        return n, x

    def _getCoFromStr_wPer(self, val):
        val_copy = val
        per_idx = val.index('|')
        
        val_copy = val_copy[:per_idx]
        N, X = self._getCoFromStr_simple(val_copy)
        
        val = val.replace('.', '').replace('|', '.')
        N2, X2 = self._getCoFromStr_simple(val)
        
        x = X2 - X
        n = N2 * N - N
        return n, x

    def _getval(self):
        return self.x / self.n


    def __cmp__(self, other):  ## comparison
        other = _convertToJfloat(other)
        if self.x == other.x and self.n == other.n:
            return int(0)

        self_x, other_x, lcm = self._prep4op1(other)

        if self.x > other.x:
            return 1
        return -1

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __le__(self, other):
        return self.__cmp__(other) <= 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0

    def __ge__(self, other):
        return self.__cmp__(other) >= 0

    def __gt__(self, other):
        return self.__cmp__(other) > 0
            
##    def __pos__(self)

    def __neg__(self): # negative object
        return Jfloat(-self.x, self.n)

    def __abs__(self): # abs value
        return Jfloat(abs(self.x), abs(self.n))

    def __invert__(self):
        return Jfloat(self.n, self.x)

##    def __invert__(self):

    def __round__(self, n): # rounds value
        return round(self._getval())

    def __floor__(self): ## return nearest in touwnwards
        return self.x // self.n

    def __ciel__(self):
        if self.x % self.n == 0:
            return self.__floor__()
        return self.__floot__() + 1

    def __trunc__(self):
        if self < 0:
            return __ciel__()
        return __floot__()
    
        

##__________________________________________________
    


    def _prep4op1(self, other):
        other = _convertToJfloat(other)
        
        lcm = Jmath.lcm(self.n, other.n)
        x = self.x
        X = other.x
        x = x * lcm // self.n
        X = X * lcm // other.n
        return x, X, lcm

    def _prep4op2(self, other):
        other = _convertToJfloat(other)
        
        x = self.x
        X = other.x
        n = self.n
        N = other.n
        return x, X, n, N

    def __add__(self, other):
        x, X, lcm = self._prep4op1(other)
        return Jfloat(X + x, lcm)
  

    def __sub__(self, other):
        x, X, lcm = self._prep4op1(other)
        return Jfloat(x - X, lcm)
        

    def __mul__(self, other):
        x, X, n, N = self._prep4op2(other)
        return Jfloat(x * X, n * N)

    def __floordiv__(self, other):
        num = self.__truediv__(other)
        return num.__floor__()

    def __truediv__(self, other):
        x, X, n, N = self._prep4op2(other)
        return Jfloat(x * N, X * n)

    # truediv

    def __mod__(self, other):
        return self - self.__floordiv__(other) * other

    def __exp__(self, other):
        x, X, n, N = self._prep4op2(other)
        if N != 1:
            raise ValueError
        return Jfloat(x ** X, n ** x)

## _________________________________________________________

    def __radd__(self, other):
        return self.__add__(other)

    def __rsub__(self, other):
        x, X, lcm = self._prep4op1(other)
        return Jfloat(X - x, lcm)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __rfloordiv__(self, other):
        num = self.__rtruediv__(other)
        return num.__floor__()

    def __rtruediv__(self, other):
        X, x, N, n = self._prep4op2(other)
        return Jfloat(x * N, X * n)

    def __rmod__(self, other):
        return other - self.__rmul__(self.__rfloordiv__(other))

    def __rexp__(self, other):
        X, x, N, n = self._prep4op2(other)
        if N != 1:
            raise ValueError
        return Jfloat(x ** X, n ** x)

## ________________________________________________________

    def __iadd__(self, other):
        return self.__add__(other)

    def __isub__(self, other):
        return self.__sub__(other)

    def __imul__(self, other):
        return self.__mul__(other)

    def __ifloordiv__(self, other):
        return self.__floordiv__(other)

    def __itruediv__(self, other):
        return self.__truediv__(other)

    def __imod__(self, other):
        return self.__mod__(other)

    def __iexp__(self, other):
        return self.__mod__(other)

## _____________________________________________________

    def __float__(self):
        return self.x / self.n

    def __int__(self):
        return int(self.__float__())

    def __Jfloat__(self):
        return self

## ______________________________________________________

    def __str__(self):
        return str(self.x) + ' / ' + str(self.n)

    def __unicode__(self):
        return self.__str__()

    def __format__(self, _):
        return self.__str__()

    def __repr__(self):
        return self.__str__()

    def __dir__(self):
        return self.x, self.n

    def __bool__(self):
        return self.x != 0

## _____________________________________________________

    def __getattr__(self, name):
        raise AttributeError (str(name))

## _____________________________________________________

    def __copy__(self):
        return Jfloat(self.x, self.n)

    def __deepcopy__(self):
        return self

## _____________________________________________________

    def __getstate__(self):
        return self.x, self.n

    def __setstate__(self, state):
        self.x, self.n = state


def _convertToJfloat(num):
    if isinstance(num, Jfloat):
        return num

    try:
        return num.__Jfloat__()
    except:
        return Jfloat(num)

